from urlparse import urlunsplit, urljoin, uses_relative
from cgi import parse_qsl
from mext.dispatch import *
from mext.const import *

from mext.uri.parse import urlsplit
from mext.uri.encode import dict_to_query
from mext.uri.schemes import *

__all__ = ['Uri', 'RelativeUri', 'AbsoluteUri', 'urlsplit']

OTHER = const()


class Uri(object):
    adapt = Adapter('__uri__')

    def __init__(self, *args):
        if len(args) == 1:
            args = urlsplit(unicode(args[0]))
        self.scheme, self.netloc, self.path, self.query, self.fragment = args
        cls = AbsoluteUri if self.scheme else RelativeUri
        if not isinstance(self, cls):
            if not issubclass(cls, self.__class__):
                raise TypeError("%r is not a subclass of %r" % (cls, self.__class__))
            self.__class__ = cls
        self._post_init()

    def _post_init(self):
        pass

    @property
    def parts(self):
        return (self.scheme, self.netloc, self.path, self.query, self.fragment)

    def __str__(self):
        return urlunsplit(self.parts)

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % repr(str(self))

    @property
    def path_with_query(self):
        path = self.path
        if self.query:
            path += '?' + self.query
        return path

    @classmethod
    def from_parts(cls, scheme=None, netloc=None, path=None, query=None, fragment=None):
        return cls(scheme, netloc, path, query, fragment)

    def copy(self):
        return self.__class__(str(self))

    def join(self, ref):
        return Uri(urljoin(str(self), str(ref)))

    #@@ maybe just use .join('#frag') ?
    def with_fragment(self, fragment):
        return Uri(self.scheme, self.netloc, self.path, self.query, fragment)

    def as_dir(self, save_query=False):
        r = self.copy()
        if not r.path.endswith('/'):
            r.path += '/'
        if not save_query:
            r.query = None
            r.fragment = None
        return r

    def defrag(self):
        if self.fragment:
            return self.with_fragment(None)
        else:
            return self


    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == other

    def __ne__(self, other):
        return not (self == other)


    def _get_query(self):
        if self._query is OTHER:
            self._query = dict_to_query(self._query_dict)
            self._query_dict = OTHER
        return self._query
    def _set_query(self, query):
        self._query = query
        self._query_dict = OTHER
    query = property(_get_query, _set_query)


    def _get_query_dict(self):
        if self._query_dict is OTHER:
            self._query_dict = dict(parse_qsl(self._query)) if self._query else {}
            self._query = OTHER
        return self._query_dict
    def _set_query_dict(self, query_dict):
        self._query = OTHER
        self._query_dict = query_dict
    query_dict = property(_get_query_dict, _set_query_dict)




Uri.adapt[basestring] = Uri
Uri.adapt.register('urllib2.Request', lambda req: AbsoluteUri(req.get_full_url()))

class AbsoluteUri(Uri):
    is_absolute = True
    def relative(self, base):
        base = Uri.adapt(base)
        if not base.is_absolute:
            raise ValueError("Cannon generate relative links from a relative base")
        elif (base.scheme or self.scheme) not in uses_relative:
            return self
        elif self.scheme == base.scheme and self.netloc == base.netloc:
            parts = ['', '', self.path, self.query, self.fragment]
            if base.path == self.path and base.query == self.query:
                parts[2] = parts[3] = ''
            elif '/' in self.path:
                #@@ better relative paths
                #@@ urlparse.non_hierarchical
                base_split = base.path.rsplit('/', 1)
                self_split = self.path.rsplit('/', 1)
                if base_split[0] == self_split[0]:
                    parts[2] = self_split[1]
            return RelativeUri(*parts)
            #assert base.join(ref) == self, repr((base, ref, base.join(ref), self))
        else:
            return self

    def _post_init(self):
        if self.fragment and self.scheme not in scheme_uses_fragment:
            raise ValueError('URI scheme %r does not support fragment identifiers' % self.scheme)
        if not self.path and self.scheme in scheme_requires_path:
            self.path = u'/'
        super(AbsoluteUri, self)._post_init()


class RelativeUri(Uri):
    is_absolute = False

    def __str__(self):
        if (self.scheme or self.netloc or self.path or self.query):
            return Uri.__str__(self)
        else:
            return '#%s' % (self.fragment or '')

    def __nonzero__(self):
        return bool(self.scheme or self.netloc or self.path or self.query or self.fragment)











