"""
    URL path / query generation
"""
import string, codecs

__all__ = ['urlquote', 'dict_to_query']

## def _urlquote(*args, **params):
##     path = u'/'.join(args).encode('utf8')
##     path = quote_plus(path, '/')
##     if not path.startswith('/'):
##         path = '/' + path
##     query = encode_query(**params)
##     return (path + '?' + query) if query else path
##

## def encode_query(*args, **params):
##     query = list(args)
##     for key, value in params.iteritems():
##         value_str = quote_plus(unicode(value).encode('utf-8'))
##         query.append("%s=%s" % (key, value_str))
##     return '&'.join(query)








def _new_quote_table(base, safe_chars, unsafe_chars=''):
    r = base[:]
    _alter_safe(r, safe_chars)
    _alter_unsafe(r, unsafe_chars)
    return r

def _alter_safe(table, safe_chars):
    for c in safe_chars:
        if isinstance(c, str):
            i, c = ord(c), unicode(c)
        else:
            i, c = c, unicode(chr(c))
        table[i] = unicode(c)

def _alter_unsafe(table, unsafe_chars):
    for c in unsafe_chars:
        if isinstance(c, str):
            c = ord(c)
        table[c] = _qt_full[c]


# table to quote ALL chars
_qt_full = map(u'%%%02X'.__mod__, xrange(256))

# table to quote all chars but the alphanumerics and chars allowed
# in every scheme according to RFC 1738
_qt_safe = _new_quote_table(_qt_full, "-_.~")
#_qt_safe = _new_quote_table(_qt_full, "$-_.+!*'(),")
_alter_safe(_qt_safe, string.ascii_letters)
_alter_safe(_qt_safe, string.digits)

# table to quote all chars possibly allowed in any specific scheme
# if any chars are reserved in a scheme, you will take care when
# encoding it's components.
#_qt_safe2 = _new_quote_table(_qt_safe, ';/?:@=&')
_qt_safe2 = _new_quote_table(_qt_safe, "!*'();:@&=+$,/?%#[]")


# For example, to encode a path component of an HTTP URL,
# you can use _qt_safe2 but with '?' marked as unsafe
#
# Technically '/;?' must be marked as unsafe, but normally
# the path is already encoded with those chars in their real meaning.
#
# When decoding the URI, you need to do it per-component


_qt_cache = {
    # when encoding something witout knowledge of reserved chars,
    # use urlquote(url, None)
    None: _qt_safe,
    # if you know that there are no reserved chars in your scheme,
    # use urlencode(url, '')
    '': _qt_safe2
}




def urlquote(s, unsafe=None):
    if isinstance(s, unicode):
        s = s.encode('utf8')
    try:
        qtable = _qt_cache[unsafe]
    except KeyError:
        qtable = _qt_safe2[:]
        _alter_unsafe(qtable, unsafe)
        _qt_cache[unsafe] = qtable
    return str(codecs.charmap_decode(s, 'strict', qtable)[0])

def _q_path(s):
    return urlquote(s, '?')


# escape name or value of a single http query argument
_qt_qval = _new_quote_table(_qt_safe2, '', '&=+')
_qt_qval[ord(' ')] = u'+'

## class DebugMap(object):
##     def __getitem__(self, key):
##         r = _qt_qval[key]
##         print 'key:', repr(key)
##         print 'res:', repr(r)
##         #return key
##         return r
##
## #print '!', `_qt_qval[83]`
## _qt_qval_debug = DebugMap()
##
##
def _qval(s):
    if isinstance(s, unicode):
        s = s.encode('utf8')
    #print `s`
    return str(codecs.charmap_decode(s, 'strict', _qt_qval)[0])


#@@ multiple values per key
def dict_to_query(query):
    return '&'.join((_qval(k)+'='+_qval(v)) for (k,v) in query.iteritems())

##
## print _qval('a')
## print _qval(u'a')
## print `_qval('\xFF')`
## #print dict_to_query({'a': 'foo&foo', 'b': 'bar=x'})
##
## t = '\xd0\xba\xd0\xbe\xd0\xb4\xd0\xb8\xd0\xba \xd1\x8e\xd0\xbd\xd0\xb8\xd0\xba\xd0\xbe\xd0\xb4\xd0\xb8\xd0\xba\n\n--\nVersion: 1.0 RC1 dev\nPlatform: Windows-XP-5.1.2600-SP3'
## print `_qval(t)`
##
